<div style="{{ styles.container }}">
  <div style="{{ styles.headerSuccess }}">
    Deploying <span style="{{ styles.code }}">{{ deploymentName }}</span>
  </div>

  {% include 'runtime-notes.html.j2'%}

  <div style="{{ styles.base }}">Uploading dependencies...</div>
</div>
