<div style="{{ styles.container }}">
  <div style="{{ styles.headerSuccess }}">Connect to Modelbit</div>
  <div style="{{ styles.bottomSpacing }}">
    Open <a style="{{ styles.link }}" href="{{ linkUrl }}" target="_blank">{{ displayUrl }}</a> to authenticate this kernel.
    <a style="{{ styles.link }}" href="https://doc.modelbit.com/" target="_blank">Learn more.</a>
  </div>
  {% include 'pip-upgrade.html.j2'%}
</div>
