<div style="{{ styles.container }}">
  <div style="{{ styles.base }}">
    You're <span style="{{ styles.connected }}">connected</span> to Modelbit as {{ email }}.
    Workspace: <span style="{{ styles.info }}">{{ workspace }}</span>.
    {% if inRegion %}
      Region: <span style="{{ styles.info }}">{{ inRegion }}</span>
    {% endif %}
  </div>
  {% include 'pip-upgrade.html.j2'%}
  {% include 'list-warnings.html.j2'%}
</div>
